#include <QApplication>
#include <QWebEngineScriptCollection>
#include <QWebEngineProfile>
#include <QWebEngineSettings>
#include <QWebEngineView>
#include <QTimer>
#include <QDebug>
#include <QLabel>

QString toDataUrl(QString html)
{
    return QString::fromLatin1("data:text/html,") + html;
}

QString newPage()
{
    QString value = QString::fromLatin1(
                "\
                <html><head><title>Original title</title></head><body>\
                <p>Test Page runscripthere</p>\
                </body></html>");
    return value.simplified();
}
QString startPage(QString s)
{
    QString value = QString::fromLatin1(
                "\
                <html><body>\
                <a onclick=\"window.open('%1');\
                             window.open('%2');\
                             return false;\" href=\"#foo\" id=\"clickme\">Click here</a>\
                <script>document.getElementById('clickme').click()</script>\
                </body></html>");

    return value.arg(s).arg(s).simplified();
}

class MyWebView;
QList<MyWebView *> &tabs()
{
    static QList<MyWebView *> list;
    return list;
}

int &run()
{
    static int run = 1;
    return run;
}

void onRunFinished();

class MyWebView : public QWebEngineView
{
    Q_OBJECT

public slots:
    void onLoadFinished(bool ok)
    {
        Q_UNUSED(ok);
        m_finished = true;

        bool all_finished = true;
        for (int i = 0; i < tabs().size(); ++i) {
            if (!tabs()[i]->m_finished) {
                all_finished = false;
                break;
            }
        }

        // When all pages finished loading, and probably finished script execution.
        if (all_finished) {
            QTimer::singleShot(1000, onRunFinished);
        }
    }
protected:
    QWebEngineView *createWindow(QWebEnginePage::WebWindowType type)
    {
        static int i = 0;
        Q_UNUSED(type);
        MyWebView *newTab = new MyWebView(this);
        newTab->move(0, 50 + i * 100);
        newTab->resize(100, 100);
        newTab->show();
        tabs().append(newTab);

        QLabel *label = new QLabel(this);
        label->setText(QStringLiteral("No Title"));
        label->resize(100, 50);
        label->move(150, 50 + i * 100);
        label->show();
        QObject::connect(newTab, &QWebEngineView::destroyed, label, &QLabel::deleteLater);
        QObject::connect(newTab, &QWebEngineView::titleChanged, label, &QLabel::setText);

        ++i;
        return newTab;
    }
public:
    MyWebView(QWidget *parent = nullptr) : QWebEngineView(parent)
    {
        QObject::connect(this, &QWebEngineView::loadFinished, this, &MyWebView::onLoadFinished);
    }

    bool m_finished = false;
};

void onRunFinished()
{
    bool all_titles_changed = true;

    // Check if all titles have been changed by the custom userscript.
    for (int i = 0; i < tabs().size(); ++i) {
        qDebug() << "run " << run() << " title: " << tabs()[i]->title();
        if (tabs()[i]->title() != QStringLiteral("userscripts work")) {
            all_titles_changed = false;
            break;
        }
    }

    if (all_titles_changed) {
        qDebug() << "run " << run() << ": userscripts ran on all tabs";
    } else {
        qDebug() << "run " << run() << ": userscripts didn't run on all tabs";
    }


    // Do another run by removing all pages except first one, and reloading initial content.
    if (run() == 1) {
        run() += 1;

        for (int i = 1; i < tabs().size(); ++i) {
            tabs()[i]->deleteLater();
        }

        MyWebView *first = tabs()[0];
        tabs().clear();
        tabs().append(first);
        first->reload();
    }

    if (run() >= 2)
        qDebug() << "Finished";
}

int main(int argc, char *argv[])
{
    QApplication a(argc, argv);

    // Set tup JS window.open to work().
    QWebEngineProfile *defaultProfile = QWebEngineProfile::defaultProfile();
    defaultProfile->settings()->setAttribute(QWebEngineSettings::JavascriptCanOpenWindows, true);
    defaultProfile->settings()->setAttribute(QWebEngineSettings::JavascriptEnabled, true);

    // Userscript to set the title after document load.
    QWebEngineScript script;
    script.setInjectionPoint(QWebEngineScript::Deferred);
    script.setSourceCode(QString::fromLatin1(\
        "console.error('User script executed:'); document.title = 'userscripts work';"));
    defaultProfile->scripts()->insert(script);

    // Show initial view.
    MyWebView *view = new MyWebView();
    view->show();
    tabs().append(view);

    // Set initial view content.
    QString startPageContent = startPage(toDataUrl(newPage()));
    QUrl initialUrl = QUrl::fromUserInput(toDataUrl(startPageContent));
    view->load(initialUrl);

    // Handle destruction.
    int result = a.exec();
    Q_UNUSED(result);
    delete view;
}

#include "main.moc"
